/************************************************************************

(C) Copyright 2004 - Analog Devices, Inc.  All rights reserved.

File Name:		AppCompress.c

Author:			Bob Nakib

Date Modified:	08/11/04

Software:       VisualDSP++3.5

Hardware:		ADSP-BF533/532/531

Special Connections:  None

Purpose:		Compresses a .LDR build that contains a user-application
				with Init code. The compressed application code will be 
				returned as	CompressedApp.ldr, in the same folder as the
				original user-application file.

*************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

int main(int argc, char *argv[])
{
	FILE *fp; // file pointer used for I/O
	unsigned char *buf_in, *buf_out; // source and destination buffers for compress
	unsigned char *app_addr; // application start address, used to pass to compress function
	int init_dxecount, app_dxecount, result, j; // Init/app code sizes, and the result of the compression
	unsigned long app_len, comp_len, i, fileSize=0; // app/compressed total code sizes (w/ headers)
	char *pathname, *newpath;
	
	union union1  		// used to encode uncompress parameters into the output file
	{
		unsigned long uncompr_len;
		unsigned char c1[4];
	} u1;
	
	union union2
	{
		unsigned long source_len;
		unsigned char c2[4];
	} u2;	

	// PARSE COMMAND-LINE ARGUMENTS
	
	if((argc != 3) || (strcmp(argv[2],"0") && strcmp(argv[2],"1") && strcmp(argv[2],"9") && strcmp(argv[2],"-1")))
	{
   		printf("Usage: compress.exe PATH COMPRESSION_LEVEL");
        fflush(0);
        system("PAUSE");
        exit(1);
    }
    else
    {
        pathname = malloc(strlen(argv[1])*sizeof(char)); // allocate memory for filepath
        
       	if(!pathname) // check for allocation errors
       	{
           	printf("Memory allocation failed.\n");
            fflush(0);
            system("PAUSE");
    		exit(1);
		}
        
        strcpy(pathname, argv[1]);
    }

    // OPEN THE FILE TO CALCULATE SIZE
	
	if ((fp = fopen(pathname, "r")) == NULL)   // open source .LDR file for read
	{
		printf("Error opening file, check the path.\n");
        fflush(0);
        system("PAUSE");
		exit(1);
	}

	while(!feof(fp))        // parse the file to determine how much memory to allocate
	{
		fgetc(fp);

		if(ferror(fp))
		{
			printf("Error reading source file.\n");
            fflush(0);
            system("PAUSE");
			exit(1);
		}

		fileSize++;
	}

	if(fclose(fp)==EOF)            // close file read buffer
	{
		printf("Error closing source file.\n");
        fflush(0);
        system("PAUSE");
		exit(1);
	}
	
	printf("\n1- File open, size calculated. fileSize = %lu\n", fileSize);
    fflush(0);
	
	// ALLOCATE APPROPRIATE SPACE ON THE HEAP

	buf_in = malloc(fileSize*sizeof(char));  // allocate appropriate memory for source file

	if(!buf_in) // check for allocation errors
	{
		printf("Memory allocation failed.\n");
        fflush(0);
        system("PAUSE");
		exit(1);
	}

	printf("2- Space allocated on heap. Buffer address = %d\n", &buf_in[0]);
    fflush(0);
	
	// OPEN THE FILE AND READ ONTO THE HEAP

	if ((fp = fopen(pathname, "r")) == NULL)  // open source .LDR file for copy
	{
		printf("Error opening file, check the path.\n");
        fflush(0);
        system("PAUSE");
		exit(1);
	}
		
	for(i=0;(fscanf(fp, "%x", &buf_in[i])) != EOF; i++);  // copy source file into buf_in buffer

	if(fclose(fp)==EOF)  // close file read buffer
	{
		printf("Error closing source file.\n");
        fflush(0);
        system("PAUSE");
		exit(1);
	}
	
    printf("3- File read onto heap. Bytes used = %lu\n", i);
    fflush(0);

	// FIND THE APPLICATION START ADDRESS
	
	init_dxecount = buf_in[10] | (buf_in[11]<<8) | (buf_in[12]<<16) | (buf_in[13]<<24);  // read the Init code size
	
	app_addr = &buf_in[14+init_dxecount]; // get the application start address
	
	printf("4- Application start address found. Init size = %d, Address = %d\n", (init_dxecount+14), app_addr);
    fflush(0);
	
	// FIND THE APPLICATION SIZE AND SAVE START OF COMPRESSED CODE
	
	app_dxecount = buf_in[14+init_dxecount+10] | (buf_in[14+init_dxecount+11]<<8) // read the application code size
                  | (buf_in[14+init_dxecount+12]<<16) | (buf_in[14+init_dxecount+13]<<24);

	app_len = app_dxecount + 14; // total application size (w/ headers)
	
    /*  The destination buffer should be 0.1% larger than the source buffer plus 12 bytes 
        because of overhead considerations. Furthermore, note that comp_len will be 
        overwritten with the compressed code length after executing compress2() */
    comp_len = (ceil(app_len*1.001)+12);

    buf_out = malloc(app_len*sizeof(char));        // allocate a buffer for output of compression
    
	if(!buf_out) // check for allocation errors
	{
		printf("Memory allocation failed.\n");
        fflush(0);
        system("PAUSE");
		exit(1);
	}    
	
	printf("5- Application size found, output buffer set, start of compress.\nApp size = %lu,\
  Comp buffer size = %lu, Out buffer start address = %d\n", app_len, comp_len, &buf_out[0]);
    fflush(0);

	// COMPRESS THE APPLICATION AND GET THE COMPRESSED CODE LENGTH
	
    result = compress2(buf_out, &comp_len, app_addr, app_len, atoi(argv[2]));

    printf("6- Application compressed, RESULT = %d, Comp buffer size = %lu\n", result, comp_len);
    fflush(0);
    
	// CREATE THE NEW .LDR FILE

	for (j=(strlen(pathname)-1); j>0; j--)  // parse the path and prepare new filepath
	{
		if((pathname[j] == '/') || (pathname[j] == '\\')) break;
   	}

   	newpath = malloc((j+2)*sizeof(char)); // allocate memory for new filepath
     	
   	if(!newpath) // check for allocation errors
	{
		printf("Memory allocation failed.\n");
        fflush(0);
        system("PAUSE");
		exit(1);
    }
            
    strncpy(newpath,pathname,(j+1));  // copy the containing directory
    newpath[j+1]='\0';                // string null-termination

   	newpath = realloc(newpath, (j+18));    // reallocate for new file name
    
    if(!newpath) // check for allocation errors
    {
		printf("Memory allocation failed.\n");
        fflush(0);
        system("PAUSE");
		exit(1);
    }
            
	strcat(newpath,"CompressedApp.ldr");  // add new filename to the path

	printf("7- New path set: %s\n", newpath);
	fflush(0);
	
    // WRITE THE NEW .LDR FILE WITH THE COMPRESSED CODE

    if ((fp = fopen(newpath, "w")) == NULL) // open destination .LDR file for write
	{
		printf("Error creating destination file.\n");
        fflush(0);
        system("PAUSE");
		exit(1);
	}
	
	for(i=0;i<(14+init_dxecount);i++)  // write untouched Init code to destination file
    {                                  // must pad with 00's since only 16-bit reads from 
                                       // Flash are possible
        fprintf(fp, "%.2x\n", buf_in[i]);
        fprintf(fp, "00\n");
    }
 	
 	u1.uncompr_len = app_len;
	u2.source_len = comp_len;

    for(i=0;i<4;i++)
    {	
        fprintf(fp, "%.2x\n", u1.c1[i]);    // encode uncompr_len into the file to
        fprintf(fp, "00\n");                // be used by uncompress()
    }

    for(i=0;i<4;i++)
    {	
        fprintf(fp, "%.2x\n", u2.c2[i]);    // encode source_len into the file to
        fprintf(fp, "00\n");                // be used by uncompress()
    }

	for(i=0;i<comp_len;i++) fprintf(fp, "%.2x\n", buf_out[i]); // append compressed code to file

	if(fclose(fp)==EOF)  // close file write buffer
	{
		printf("Error closing destination file.\n");
        fflush(0);
        system("PAUSE");
		exit(1);
	}
	
	printf("8- New file written.\n");
    fflush(0);
		
    system("PAUSE");
    return 0;
}
